import re
from dataclasses import dataclass
from enum import Enum
from typing import Any, Optional

from dacite import from_dict, Config

from roborock.code_mappings import RoborockDockWashTowelModeCode, RoborockDockTypeCode, RoborockMopIntensityCode, \
    RoborockStateCode
from .code_mappings import RoborockMopModeCode, RoborockDockErrorCode, \
    RoborockErrorCode, RoborockDockDustCollectionModeCode, RoborockFanPowerCode


def to_snake(s):
    return re.sub('([A-Z]+)', '_\\1', s).lower()


def decamelize(d):
    if isinstance(d, list):
        return [decamelize(i) if isinstance(i, (dict, list)) else i for i in d]
    return {to_snake(a): decamelize(b) if isinstance(b, (dict, list)) else b for a, b in d.items()}


@dataclass
class RoborockBase:

    @classmethod
    def from_dict(cls, data: dict[str, any]):
        return from_dict(cls, decamelize(data), config=Config(cast=[Enum]))

    def as_dict(self):
        return self.__dict__


@dataclass
class Reference(RoborockBase):
    r: Optional[str] = None
    a: Optional[str] = None
    m: Optional[str] = None
    l: Optional[str] = None


@dataclass
class RRiot(RoborockBase):
    u: Optional[str] = None
    s: Optional[str] = None
    h: Optional[str] = None
    k: Optional[str] = None
    r: Optional[Reference] = None


@dataclass
class UserData(RoborockBase):
    uid: Optional[int] = None
    tokentype: Optional[str] = None
    token: Optional[str] = None
    rruid: Optional[str] = None
    region: Optional[str] = None
    countrycode: Optional[str] = None
    country: Optional[str] = None
    nickname: Optional[str] = None
    rriot: Optional[RRiot] = None
    tuya_device_state: Optional[int] = None
    avatarurl: Optional[str] = None


@dataclass
class HomeDataProductSchema(RoborockBase):
    id: Optional[Any] = None
    name: Optional[Any] = None
    code: Optional[Any] = None
    mode: Optional[Any] = None
    type: Optional[Any] = None
    product_property: Optional[Any] = None
    desc: Optional[Any] = None


@dataclass
class HomeDataProduct(RoborockBase):
    id: Optional[str] = None
    name: Optional[str] = None
    code: Optional[str] = None
    model: Optional[str] = None
    iconurl: Optional[str] = None
    attribute: Optional[Any] = None
    capability: Optional[int] = None
    category: Optional[str] = None
    schema: Optional[list[HomeDataProductSchema]] = None


@dataclass
class HomeDataDeviceStatus(RoborockBase):
    id: Optional[Any] = None
    name: Optional[Any] = None
    code: Optional[Any] = None
    model: Optional[Any] = None
    icon_url: Optional[Any] = None
    attribute: Optional[Any] = None
    capability: Optional[Any] = None
    category: Optional[Any] = None
    schema: Optional[Any] = None


@dataclass
class HomeDataDevice(RoborockBase):
    duid: str
    name: str
    local_key: str
    fv: str
    attribute: Optional[Any] = None
    active_time: Optional[int] = None
    runtime_env: Optional[Any] = None
    time_zone_id: Optional[str] = None
    icon_url: Optional[str] = None
    product_id: Optional[str] = None
    lon: Optional[Any] = None
    lat: Optional[Any] = None
    share: Optional[Any] = None
    share_time: Optional[Any] = None
    online: Optional[bool] = None
    pv: Optional[str] = None
    room_id: Optional[Any] = None
    tuya_uuid: Optional[Any] = None
    tuya_migrated: Optional[bool] = None
    extra: Optional[Any] = None
    sn: Optional[str] = None
    feature_set: Optional[str] = None
    new_feature_set: Optional[str] = None
    device_status: Optional[HomeDataDeviceStatus] = None
    silent_ota_switch: Optional[bool] = None


@dataclass
class HomeDataRoom(RoborockBase):
    id: Optional[Any] = None
    name: Optional[Any] = None


@dataclass
class HomeData(RoborockBase):
    id: Optional[int] = None
    name: Optional[str] = None
    lon: Optional[Any] = None
    lat: Optional[Any] = None
    geo_name: Optional[Any] = None
    products: Optional[list[HomeDataProduct]] = None
    devices: Optional[list[HomeDataDevice]] = None
    received_devices: Optional[list[HomeDataDevice]] = None
    rooms: Optional[list[HomeDataRoom]] = None


@dataclass
class LoginData(RoborockBase):
    user_data: UserData
    home_data: HomeData
    email: Optional[str] = None


@dataclass
class Status(RoborockBase):
    msg_ver: Optional[int] = None
    msg_seq: Optional[int] = None
    state: Optional[RoborockStateCode] = None
    battery: Optional[int] = None
    clean_time: Optional[int] = None
    clean_area: Optional[int] = None
    error_code: Optional[RoborockErrorCode] = None
    map_present: Optional[int] = None
    in_cleaning: Optional[int] = None
    in_returning: Optional[int] = None
    in_fresh_state: Optional[int] = None
    lab_status: Optional[int] = None
    water_box_status: Optional[int] = None
    back_type: Optional[int] = None
    wash_phase: Optional[int] = None
    wash_ready: Optional[int] = None
    fan_power: Optional[RoborockFanPowerCode] = None
    dnd_enabled: Optional[int] = None
    map_status: Optional[int] = None
    is_locating: Optional[int] = None
    lock_status: Optional[int] = None
    water_box_mode: Optional[RoborockMopIntensityCode] = None
    mop_intensity: Optional[str] = None
    water_box_carriage_status: Optional[int] = None
    mop_forbidden_enable: Optional[int] = None
    camera_status: Optional[int] = None
    is_exploring: Optional[int] = None
    home_sec_status: Optional[int] = None
    home_sec_enable_password: Optional[int] = None
    adbumper_status: Optional[list[int]] = None
    water_shortage_status: Optional[int] = None
    dock_type: Optional[RoborockDockTypeCode] = None
    dust_collection_status: Optional[int] = None
    auto_dust_collection: Optional[int] = None
    avoid_count: Optional[int] = None
    mop_mode: Optional[RoborockMopModeCode] = None
    debug_mode: Optional[int] = None
    collision_avoid_status: Optional[int] = None
    switch_map_mode: Optional[int] = None
    dock_error_status: Optional[RoborockDockErrorCode] = None
    charge_status: Optional[int] = None
    unsave_map_reason: Optional[int] = None
    unsave_map_flag: Optional[int] = None


@dataclass
class DNDTimer(RoborockBase):
    start_hour: Optional[int] = None
    start_minute: Optional[int] = None
    end_hour: Optional[int] = None
    end_minute: Optional[int] = None
    enabled: Optional[int] = None


@dataclass
class CleanSummary(RoborockBase):
    clean_time: Optional[int] = None
    clean_area: Optional[int] = None
    clean_count: Optional[int] = None
    dust_collection_count: Optional[int] = None
    records: Optional[list[int]] = None


@dataclass
class CleanRecord(RoborockBase):
    begin: Optional[int] = None
    end: Optional[int] = None
    duration: Optional[int] = None
    area: Optional[int] = None
    error: Optional[int] = None
    complete: Optional[int] = None
    start_type: Optional[int] = None
    clean_type: Optional[int] = None
    finish_reason: Optional[int] = None
    dust_collection_status: Optional[int] = None
    avoid_count: Optional[int] = None
    wash_count: Optional[int] = None
    map_flag: Optional[int] = None


@dataclass
class Consumable(RoborockBase):
    main_brush_work_time: Optional[int] = None
    side_brush_work_time: Optional[int] = None
    filter_work_time: Optional[int] = None
    filter_element_work_time: Optional[int] = None
    sensor_dirty_time: Optional[int] = None
    strainer_work_times: Optional[int] = None
    dust_collection_work_times: Optional[int] = None
    cleaning_brush_work_times: Optional[int] = None


@dataclass
class MultiMapsListMapInfoBakMaps(RoborockBase):
    mapflag: Optional[Any] = None
    add_time: Optional[Any] = None


@dataclass
class MultiMapsListMapInfo(RoborockBase):
    mapflag: Optional[Any] = None
    add_time: Optional[Any] = None
    length: Optional[Any] = None
    name: Optional[Any] = None
    bak_maps: Optional[list[MultiMapsListMapInfoBakMaps]] = None


@dataclass
class MultiMapsList(RoborockBase):
    max_multi_map: Optional[int] = None
    max_bak_map: Optional[int] = None
    multi_map_count: Optional[int] = None
    map_info: Optional[list[MultiMapsListMapInfo]] = None


@dataclass
class SmartWashParams(RoborockBase):
    smart_wash: Optional[int] = None
    wash_interval: Optional[int] = None


@dataclass
class DustCollectionMode(RoborockBase):
    mode: Optional[RoborockDockDustCollectionModeCode] = None


@dataclass
class WashTowelMode(RoborockBase):
    wash_mode: Optional[RoborockDockWashTowelModeCode] = None


@dataclass
class NetworkInfo(RoborockBase):
    ip: str
    ssid: Optional[str] = None
    mac: Optional[str] = None
    bssid: Optional[str] = None
    rssi: Optional[int] = None


@dataclass
class RoborockDeviceInfo(RoborockBase):
    device: HomeDataDevice


@dataclass
class RoborockLocalDeviceInfo(RoborockDeviceInfo):
    network_info: NetworkInfo



import asyncio
import logging

from roborock.containers import HomeDataDevice, HomeDataDeviceField, HomeDataProduct, HomeDataProductField, NetworkInfo, \
    NetworkInfoField
from roborock.containers import RoborockLocalDeviceInfo
from roborock.local_api import RoborockLocalClient

local_ip = "<local_ip>"
local_key = "<local_key>"
device_id = "<device_id>"


async def main():
    logging_config = {
        "level": logging.DEBUG
    }
    logging.basicConfig(**logging_config)
    localdevices_info = RoborockLocalDeviceInfo({
        "device": HomeDataDevice({
            HomeDataDeviceField.NAME: "test_name",
            HomeDataDeviceField.DUID: device_id,
            HomeDataDeviceField.FV: "1",
            HomeDataDeviceField.LOCAL_KEY: local_key
        }),
        "product": HomeDataProduct({
            HomeDataProductField.MODEL: "test_model"
        }),
        "network_info": NetworkInfo({
            NetworkInfoField.IP: local_ip
        })})
    client = RoborockLocalClient({
        device_id: localdevices_info
    })
    await client.async_connect()
    props = await client.get_prop(device_id)
    print(props)


if __name__ == "__main__":
    asyncio.run(main())



from __future__ import annotations

from asyncio import Queue
from typing import Any

import async_timeout

from .exceptions import VacuumError


class RoborockQueue(Queue):
    def __init__(self, protocol: int, *args):
        super().__init__(*args)
        self.protocol = protocol

    async def async_put(self, item: tuple[Any, VacuumError | None], timeout: float | int) -> None:
        async with async_timeout.timeout(timeout):
            await self.put(item)

    async def async_get(self, timeout: float | int) -> tuple[Any, VacuumError | None]:
        async with async_timeout.timeout(timeout):
            return await self.get()



import json
import logging
from pathlib import Path
from typing import Any, Dict

import click

from roborock import RoborockException
from roborock.api import RoborockApiClient
from roborock.cloud_api import RoborockMqttClient
from roborock.containers import LoginData, RoborockDeviceInfo
from roborock.util import run_sync

_LOGGER = logging.getLogger(__name__)


class RoborockContext:
    roborock_file = Path("~/.roborock").expanduser()
    _login_data: LoginData = None

    def __init__(self):
        self.reload()

    def reload(self):
        if self.roborock_file.is_file():
            with open(self.roborock_file, 'r') as f:
                data = json.load(f)
                if data:
                    self._login_data = LoginData(data)

    def update(self, login_data: LoginData):
        data = json.dumps(login_data, default=vars)
        with open(self.roborock_file, 'w') as f:
            f.write(data)
        self.reload()

    def validate(self):
        if self._login_data is None:
            raise RoborockException("You must login first")

    def login_data(self):
        self.validate()
        return self._login_data


@click.option("-d", "--debug", default=False, count=True)
@click.version_option(package_name="python-roborock")
@click.group()
@click.pass_context
def cli(ctx, debug: int):
    logging_config: Dict[str, Any] = {
        "level": logging.DEBUG if debug > 0 else logging.INFO
    }
    logging.basicConfig(**logging_config)  # type: ignore
    ctx.obj = RoborockContext()


@click.command()
@click.option("--email", required=True)
@click.option("--password", required=True)
@click.pass_context
@run_sync()
async def login(ctx, email, password):
    """Login to Roborock account."""
    context: RoborockContext = ctx.obj
    try:
        context.validate()
        _LOGGER.info("Already logged in")
        return
    except RoborockException:
        pass
    client = RoborockApiClient(email)
    user_data = await client.pass_login(password)
    context.update(LoginData({"user_data": user_data, "email": email}))


async def _discover(ctx):
    context: RoborockContext = ctx.obj
    login_data = context.login_data()
    client = RoborockApiClient(login_data.email)
    home_data = await client.get_home_data(login_data.user_data)
    context.update(LoginData({**login_data, "home_data": home_data}))
    click.echo(
        f"Discovered devices {', '.join([device.name for device in home_data.devices + home_data.received_devices])}")


@click.command()
@click.pass_context
@run_sync()
async def discover(ctx):
    await _discover(ctx)


@click.command()
@click.pass_context
@run_sync()
async def list_devices(ctx):
    context: RoborockContext = ctx.obj
    login_data = context.login_data()
    if not login_data.home_data:
        await _discover(ctx)
        login_data = context.login_data()
    home_data = login_data.home_data
    click.echo(f"Known devices {', '.join([device.name for device in home_data.devices + home_data.received_devices])}")


@click.command()
@click.option("--cmd", required=True)
@click.option("--params", required=False)
@click.pass_context
@run_sync()
async def command(ctx, cmd, params):
    context: RoborockContext = ctx.obj
    login_data = context.login_data()
    if not login_data.home_data:
        await _discover(ctx)
        login_data = context.login_data()
    home_data = login_data.home_data
    device_map: dict[str, RoborockDeviceInfo] = {}
    for device in home_data.devices + home_data.received_devices:
        device_map[device.duid] = RoborockDeviceInfo(device=device)
    mqtt_client = RoborockMqttClient(login_data.user_data, device_map)
    await mqtt_client.send_command(home_data.devices[0].duid, cmd, params)
    mqtt_client.__del__()


cli.add_command(login)
cli.add_command(discover)
cli.add_command(list_devices)
cli.add_command(command)


def main():
    return cli()


if __name__ == "__main__":
    main()



from __future__ import annotations

import binascii
import hashlib
import json
import math
import struct
import time
from dataclasses import dataclass
from random import randint
from typing import Optional

from Crypto.Cipher import AES
from Crypto.Util.Padding import pad, unpad
from roborock.typing import RoborockCommand

from roborock.exceptions import RoborockException


def md5bin(message: str) -> bytes:
    md5 = hashlib.md5()
    md5.update(message.encode())
    return md5.digest()


def encode_timestamp(_timestamp: int) -> str:
    hex_value = f"{_timestamp:x}".zfill(8)
    return "".join(list(map(lambda idx: hex_value[idx], [5, 6, 3, 7, 1, 2, 0, 4])))


salt = "TXdfu$jyZ#TZHsg4"


@dataclass
class RoborockMessage:
    protocol: int
    payload: bytes
    seq: Optional[int] = randint(100000, 999999)
    prefix: Optional[bytes] = b''
    version: Optional[bytes] = b'1.0'
    random: Optional[int] = randint(10000, 99999)
    timestamp: Optional[int] = math.floor(time.time())

    def get_request_id(self) -> int | None:
        protocol = self.protocol
        if protocol in [4, 101, 102]:
            payload = json.loads(self.payload.decode())
            for data_point_number, data_point in payload.get("dps").items():
                if data_point_number in ["101", "102"]:
                    data_point_response = json.loads(data_point)
                    return data_point_response.get("id")
        return None

    def get_method(self) -> RoborockCommand | None:
        protocol = self.protocol
        if protocol in [4, 101, 102]:
            payload = json.loads(self.payload.decode())
            for data_point_number, data_point in payload.get("dps").items():
                if data_point_number in ["101", "102"]:
                    data_point_response = json.loads(data_point)
                    return data_point_response.get("method")
        return None

    def get_params(self) -> list | None:
        protocol = self.protocol
        if protocol in [4, 101, 102]:
            payload = json.loads(self.payload.decode())
            for data_point_number, data_point in payload.get("dps").items():
                if data_point_number in ["101", "102"]:
                    data_point_response = json.loads(data_point)
                    return data_point_response.get("params")
        return None


class RoborockParser:

    @staticmethod
    def encode(roborock_messages: list[RoborockMessage] | RoborockMessage, local_key: str):
        if isinstance(roborock_messages, RoborockMessage):
            roborock_messages = [roborock_messages]

        msg = b''
        for roborock_message in roborock_messages:
            aes_key = md5bin(encode_timestamp(roborock_message.timestamp) + local_key + salt)
            cipher = AES.new(aes_key, AES.MODE_ECB)
            payload = roborock_message.payload
            if payload:
                payload = pad(roborock_message.payload, AES.block_size)
            encrypted = cipher.encrypt(payload)
            encrypted_len = len(encrypted)
            _msg = struct.pack(
                f"!3sIIIHH{encrypted_len}s",
                "1.0".encode(),
                roborock_message.seq,
                roborock_message.random,
                roborock_message.timestamp,
                roborock_message.protocol,
                encrypted_len,
                encrypted
            )
            if payload:
                crc32 = binascii.crc32(_msg)
                _msg += struct.pack("!I", crc32)
            else:
                _msg += b'\x00\x00'
            msg += roborock_message.prefix + _msg
        return msg

    @staticmethod
    def decode(msg: bytes, local_key: str, index=0) -> tuple[list[RoborockMessage], bytes]:
        prefix = None
        original_index = index
        if len(msg) - index < 17:
            ## broken message
            return [], msg[original_index:]

        if msg[index + 4:index + 7] == "1.0".encode():
            prefix = msg[index:index + 4]
            index += 4
        elif msg[index:index + 3] != "1.0".encode():
            raise RoborockException(f"Unknown protocol version {msg[0:3]}")

        if len(msg) - index in [17]:
            [version, request_id, random, timestamp, protocol] = struct.unpack_from(
                "!3sIIIH", msg, index
            )
            return [RoborockMessage(
                prefix=prefix,
                version=version,
                seq=request_id,
                random=random,
                timestamp=timestamp,
                protocol=protocol,
                payload=b''
            )], b''

        if len(msg) - index < 19:
            ## broken message
            return [], msg[original_index:]

        [version, request_id, random, timestamp, protocol, payload_len] = struct.unpack_from(
            "!3sIIIHH", msg, index
        )
        index += 19

        if payload_len + index + 4 > len(msg):
            ## broken message
            return [], msg[original_index:]

        payload = b''
        if payload_len == 0:
            index += 2
        else:
            [payload, expected_crc32] = struct.unpack_from(f"!{payload_len}sI", msg, index)
            crc32 = binascii.crc32(msg[index - 19: index + payload_len])
            index += 4 + payload_len
            if crc32 != expected_crc32:
                raise RoborockException(f"Wrong CRC32 {crc32}, expected {expected_crc32}")

        if payload:
            aes_key = md5bin(encode_timestamp(timestamp) + local_key + salt)
            decipher = AES.new(aes_key, AES.MODE_ECB)
            payload = unpad(decipher.decrypt(payload), AES.block_size)

        [structs, remaining] = RoborockParser.decode(msg, local_key, index) if index < len(msg) else ([], b'')

        return [RoborockMessage(
            prefix=prefix,
            version=version,
            seq=request_id,
            random=random,
            timestamp=timestamp,
            protocol=protocol,
            payload=payload
        )] + structs, remaining



from __future__ import annotations

import asyncio
import logging
import socket
from asyncio import Lock
from typing import Callable, Coroutine

import async_timeout

from roborock.api import RoborockClient, SPECIAL_COMMANDS
from roborock.containers import RoborockLocalDeviceInfo
from roborock.exceptions import RoborockTimeout, CommandVacuumError, RoborockConnectionException, RoborockException
from roborock.roborock_message import RoborockParser, RoborockMessage
from roborock.typing import RoborockCommand, CommandInfoMap
from roborock.util import get_running_loop_or_create_one

_LOGGER = logging.getLogger(__name__)


class RoborockLocalClient(RoborockClient):

    def __init__(self, devices_info: dict[str, RoborockLocalDeviceInfo]):
        super().__init__("abc", devices_info)
        self.loop = get_running_loop_or_create_one()
        self.device_listener: dict[str, RoborockSocketListener] = {
            device_id: RoborockSocketListener(
                device_info.network_info.ip,
                device_info.device.local_key,
                self.on_message
            )
            for device_id, device_info in devices_info.items()
        }
        self._mutex = Lock()
        self._batch_structs: list[RoborockMessage] = []
        self._executing = False

    async def async_connect(self):
        await asyncio.gather(*[
            listener.connect()
            for listener in self.device_listener.values()
        ])

    async def async_disconnect(self) -> None:
        await asyncio.gather(*[listener.disconnect() for listener in self.device_listener.values()])

    def build_roborock_message(
            self, method: RoborockCommand, params: list = None
    ) -> RoborockMessage:
        secured = True if method in SPECIAL_COMMANDS else False
        request_id, timestamp, payload = self._get_payload(method, params, secured)
        _LOGGER.debug(f"id={request_id} Requesting method {method} with {params}")
        command_info = CommandInfoMap.get(method)
        if not command_info:
            raise RoborockException(f"Request {method} have unknown prefix. Can't execute in offline mode")
        prefix = CommandInfoMap.get(method).prefix
        request_protocol = 4
        return RoborockMessage(
            prefix=prefix,
            timestamp=timestamp,
            protocol=request_protocol,
            payload=payload
        )

    async def send_command(
            self, device_id: str, method: RoborockCommand, params: list = None
    ):
        roborock_message = self.build_roborock_message(method, params)
        response = (await self.send_message(device_id, roborock_message))[0]
        if isinstance(response, BaseException):
            raise response
        return response

    async def async_local_response(self, roborock_message: RoborockMessage):
        request_id = roborock_message.get_request_id()
        if request_id is not None:
            # response_protocol = 5 if roborock_message.prefix == secured_prefix else 4
            response_protocol = 4
            (response, err) = await self._async_response(request_id, response_protocol)
            if err:
                raise CommandVacuumError("", err) from err
            _LOGGER.debug(f"id={request_id} Response from {roborock_message.get_method()}: {response}")
            return response

    async def send_message(
            self, device_id: str, roborock_messages: list[RoborockMessage] | RoborockMessage
    ):
        if isinstance(roborock_messages, RoborockMessage):
            roborock_messages = [roborock_messages]
        local_key = self.devices_info[device_id].device.local_key
        msg = RoborockParser.encode(roborock_messages, local_key)
        # Send the command to the Roborock device
        listener = self.device_listener.get(device_id)
        _LOGGER.debug(f"Requesting device with {roborock_messages}")
        await listener.send_message(msg)

        return await asyncio.gather(
            *[self.async_local_response(roborock_message) for roborock_message in roborock_messages],
            return_exceptions=True)


class RoborockSocket(socket.socket):
    _closed = None

    @property
    def is_closed(self):
        return self._closed


class RoborockSocketListener:
    roborock_port = 58867

    def __init__(self, ip: str, local_key: str, on_message: Callable[[list[RoborockMessage]], Coroutine[None] | None],
                 timeout: float | int = 4):
        self.ip = ip
        self.local_key = local_key
        self.socket = RoborockSocket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.setblocking(False)
        self.loop = get_running_loop_or_create_one()
        self.on_message = on_message
        self.timeout = timeout
        self.is_connected = False
        self._mutex = Lock()
        self.remaining = b''

    async def _main_coro(self):
        while not self.socket.is_closed:
            try:
                message = await self.loop.sock_recv(self.socket, 4096)
                try:
                    if self.remaining:
                        message = self.remaining + message
                        self.remaining = b''
                    (parser_msg, remaining) = RoborockParser.decode(message, self.local_key)
                    self.remaining = remaining
                    await self.on_message(parser_msg)
                except Exception as e:
                    _LOGGER.exception(e)
            except BrokenPipeError as e:
                _LOGGER.exception(e)
                await self.disconnect()

    async def connect(self):
        async with self._mutex:
            if not self.is_connected or self.socket.is_closed:
                self.socket = RoborockSocket(socket.AF_INET, socket.SOCK_STREAM)
                self.socket.setblocking(False)
                try:
                    async with async_timeout.timeout(self.timeout):
                        _LOGGER.info(f"Connecting to {self.ip}")
                        await self.loop.sock_connect(self.socket, (self.ip, 58867))
                        self.is_connected = True
                except Exception as e:
                    await self.disconnect()
                    raise RoborockConnectionException(f"Failed connecting to {self.ip}") from e
                self.loop.create_task(self._main_coro())

    async def disconnect(self):
        self.socket.close()
        self.is_connected = False

    async def send_message(self, data: bytes):
        response = {}
        await self.connect()
        try:
            async with self._mutex:
                async with async_timeout.timeout(self.timeout):
                    await self.loop.sock_sendall(self.socket, data)
        except (asyncio.TimeoutError, asyncio.CancelledError):
            raise RoborockTimeout(
                f"Timeout after {self.timeout} seconds waiting for response"
            ) from None
        except BrokenPipeError as e:
            _LOGGER.exception(e)
            await self.disconnect()
        return response



import asyncio
import functools


def get_running_loop_or_create_one():
    try:
        loop = asyncio.get_event_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
    return loop


def run_in_executor():
    loop = get_running_loop_or_create_one()

    def decorator(func):
        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            return asyncio.run_coroutine_threadsafe(coro=func(*args, **kwargs), loop=loop)

        return wrapped

    return decorator


def run_sync():
    loop = get_running_loop_or_create_one()

    def decorator(func):
        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            return loop.run_until_complete(func(*args, **kwargs))

        return wrapped

    return decorator



"""The Roborock api."""

from __future__ import annotations

import asyncio
import base64
import gzip
import hashlib
import hmac
import json
import logging
import math
import secrets
import struct
import time
from random import randint
from typing import Any, Callable

import aiohttp
from Crypto.Cipher import AES
from Crypto.Util.Padding import unpad

from roborock.exceptions import (
    RoborockException, RoborockTimeout, VacuumError,
)
from .code_mappings import RoborockDockTypeCode
from .containers import (
    UserData,
    Status,
    CleanSummary,
    Consumable,
    DNDTimer,
    CleanRecord,
    HomeData,
    MultiMapsList,
    SmartWashParams,
    RoborockDeviceInfo,
    WashTowelMode,
    DustCollectionMode,
    NetworkInfo,

)
from .roborock_message import RoborockMessage
from .roborock_queue import RoborockQueue
from .typing import (
    RoborockDeviceProp,
    RoborockCommand,
    RoborockDockSummary
)

_LOGGER = logging.getLogger(__name__)
QUEUE_TIMEOUT = 4
MQTT_KEEPALIVE = 60
SPECIAL_COMMANDS = [
    RoborockCommand.GET_MAP_V1,
]


def md5hex(message: str) -> str:
    md5 = hashlib.md5()
    md5.update(message.encode())
    return md5.hexdigest()


class PreparedRequest:
    def __init__(self, base_url: str, base_headers: dict = None) -> None:
        self.base_url = base_url
        self.base_headers = base_headers or {}

    async def request(
            self, method: str, url: str, params=None, data=None, headers=None
    ) -> dict | list:
        _url = "/".join(s.strip("/") for s in [self.base_url, url])
        _headers = {**self.base_headers, **(headers or {})}
        async with aiohttp.ClientSession() as session:
            async with session.request(
                    method,
                    _url,
                    params=params,
                    data=data,
                    headers=_headers,
            ) as resp:
                return await resp.json()


class RoborockClient:

    def __init__(self, endpoint: str, devices_info: dict[str, RoborockDeviceInfo]) -> None:
        self.devices_info = devices_info
        self._salt = "TXdfu$jyZ#TZHsg4"
        self._endpoint = endpoint
        self._nonce = secrets.token_bytes(16)
        self._waiting_queue: dict[int, RoborockQueue] = {}
        self._status_listeners: list[Callable[[int, str], None]] = []

    def add_status_listener(self, callback: Callable[[int, str], None]):
        self._status_listeners.append(callback)

    async def async_disconnect(self) -> Any:
        raise NotImplementedError

    async def on_message(self, messages: list[RoborockMessage]) -> None:
        try:
            for data in messages:
                protocol = data.protocol
                if protocol == 102 or protocol == 4:
                    payload = json.loads(data.payload.decode())
                    for data_point_number, data_point in payload.get("dps").items():
                        if data_point_number == "102":
                            data_point_response = json.loads(data_point)
                            request_id = data_point_response.get("id")
                            queue = self._waiting_queue.get(request_id)
                            if queue:
                                if queue.protocol == protocol:
                                    error = data_point_response.get("error")
                                    if error:
                                        await queue.async_put(
                                            (
                                                None,
                                                VacuumError(
                                                    error.get("code"), error.get("message")
                                                ),
                                            ),
                                            timeout=QUEUE_TIMEOUT,
                                        )
                                    else:
                                        result = data_point_response.get("result")
                                        if isinstance(result, list) and len(result) == 1:
                                            result = result[0]
                                        await queue.async_put(
                                            (result, None), timeout=QUEUE_TIMEOUT
                                        )
                elif protocol == 301:
                    payload = data.payload[0:24]
                    [endpoint, _, request_id, _] = struct.unpack("<15sBH6s", payload)
                    if endpoint.decode().startswith(self._endpoint):
                        iv = bytes(AES.block_size)
                        decipher = AES.new(self._nonce, AES.MODE_CBC, iv)
                        decrypted = unpad(
                            decipher.decrypt(data.payload[24:]), AES.block_size
                        )
                        decrypted = gzip.decompress(decrypted)
                        queue = self._waiting_queue.get(request_id)
                        if queue:
                            if isinstance(decrypted, list):
                                decrypted = decrypted[0]
                            await queue.async_put((decrypted, None), timeout=QUEUE_TIMEOUT)
        except Exception as ex:
            _LOGGER.exception(ex)

    async def _async_response(self, request_id: int, protocol_id: int = 0) -> tuple[Any, VacuumError | None]:
        try:
            queue = RoborockQueue(protocol_id)
            self._waiting_queue[request_id] = queue
            (response, err) = await queue.async_get(QUEUE_TIMEOUT)
            return response, err
        except (asyncio.TimeoutError, asyncio.CancelledError):
            raise RoborockTimeout(
                f"Timeout after {QUEUE_TIMEOUT} seconds waiting for response"
            ) from None
        finally:
            del self._waiting_queue[request_id]

    def _get_payload(
            self, method: RoborockCommand, params: list = None, secured=False
    ):
        timestamp = math.floor(time.time())
        request_id = randint(10000, 99999)
        inner = {
            "id": request_id,
            "method": method,
            "params": params or [],
        }
        if secured:
            inner["security"] = {
                "endpoint": self._endpoint,
                "nonce": self._nonce.hex().lower(),
            }
        payload = bytes(
            json.dumps(
                {
                    "dps": {"101": json.dumps(inner, separators=(",", ":"))},
                    "t": timestamp,
                },
                separators=(",", ":"),
            ).encode()
        )
        return request_id, timestamp, payload

    async def send_command(
            self, device_id: str, method: RoborockCommand, params: list = None
    ):
        raise NotImplementedError

    async def get_status(self, device_id: str) -> Status:
        status = await self.send_command(device_id, RoborockCommand.GET_STATUS)
        if isinstance(status, dict):
            return Status.from_dict(status)

    async def get_dnd_timer(self, device_id: str) -> DNDTimer:
        try:
            dnd_timer = await self.send_command(device_id, RoborockCommand.GET_DND_TIMER)
            if isinstance(dnd_timer, dict):
                return DNDTimer.from_dict(dnd_timer)
        except RoborockTimeout as e:
            _LOGGER.error(e)

    async def get_clean_summary(self, device_id: str) -> CleanSummary:
        try:
            clean_summary = await self.send_command(
                device_id, RoborockCommand.GET_CLEAN_SUMMARY
            )
            if isinstance(clean_summary, dict):
                return CleanSummary.from_dict(clean_summary)
            elif isinstance(clean_summary, bytes):
                return CleanSummary(clean_time=int.from_bytes(clean_summary, 'big'))
        except RoborockTimeout as e:
            _LOGGER.error(e)

    async def get_clean_record(self, device_id: str, record_id: int) -> CleanRecord:
        try:
            clean_record = await self.send_command(
                device_id, RoborockCommand.GET_CLEAN_RECORD, [record_id]
            )
            if isinstance(clean_record, dict):
                return CleanRecord.from_dict(clean_record)
        except RoborockTimeout as e:
            _LOGGER.error(e)

    async def get_consumable(self, device_id: str) -> Consumable:
        try:
            consumable = await self.send_command(device_id, RoborockCommand.GET_CONSUMABLE)
            if isinstance(consumable, dict):
                return Consumable.from_dict(consumable)
        except RoborockTimeout as e:
            _LOGGER.error(e)

    async def get_wash_towel_mode(self, device_id: str) -> WashTowelMode:
        try:
            washing_mode = await self.send_command(device_id, RoborockCommand.GET_WASH_TOWEL_MODE)
            if isinstance(washing_mode, dict):
                return WashTowelMode.from_dict(washing_mode)
        except RoborockTimeout as e:
            _LOGGER.error(e)

    async def get_dust_collection_mode(self, device_id: str) -> DustCollectionMode:
        try:
            dust_collection = await self.send_command(device_id, RoborockCommand.GET_DUST_COLLECTION_MODE)
            if isinstance(dust_collection, dict):
                return DustCollectionMode.from_dict(dust_collection)
        except RoborockTimeout as e:
            _LOGGER.error(e)

    async def get_smart_wash_params(self, device_id: str) -> SmartWashParams:
        try:
            mop_wash_mode = await self.send_command(device_id, RoborockCommand.GET_SMART_WASH_PARAMS)
            if isinstance(mop_wash_mode, dict):
                return SmartWashParams.from_dict(mop_wash_mode)
        except RoborockTimeout as e:
            _LOGGER.error(e)

    async def get_dock_summary(self, device_id: str, dock_type: RoborockDockTypeCode) -> RoborockDockSummary:
        try:
            commands = [self.get_dust_collection_mode(device_id)]
            if dock_type == RoborockDockTypeCode['3']:
                commands += [self.get_wash_towel_mode(device_id), self.get_smart_wash_params(device_id)]
            [
                dust_collection_mode,
                wash_towel_mode,
                smart_wash_params
            ] = (
                        list(await asyncio.gather(*commands))
                        + [None, None]
                )[:3]

            return RoborockDockSummary(dust_collection_mode, wash_towel_mode, smart_wash_params)
        except RoborockTimeout as e:
            _LOGGER.error(e)

    async def get_prop(self, device_id: str) -> RoborockDeviceProp | None:
        [status, dnd_timer, clean_summary, consumable] = await asyncio.gather(
            *[
                self.get_status(device_id),
                self.get_dnd_timer(device_id),
                self.get_clean_summary(device_id),
                self.get_consumable(device_id),
            ]
        )
        last_clean_record = None
        if clean_summary and clean_summary.records and len(clean_summary.records) > 0:
            last_clean_record = await self.get_clean_record(
                device_id, clean_summary.records[0]
            )
        dock_summary = None
        if status and status.dock_type != RoborockDockTypeCode['0']:
            dock_summary = await self.get_dock_summary(device_id, status.dock_type)
        if any([status, dnd_timer, clean_summary, consumable]):
            return RoborockDeviceProp(
                status, dnd_timer, clean_summary, consumable, last_clean_record, dock_summary
            )
        return None

    async def get_multi_maps_list(self, device_id) -> MultiMapsList:
        try:
            multi_maps_list = await self.send_command(
                device_id, RoborockCommand.GET_MULTI_MAPS_LIST
            )
            if isinstance(multi_maps_list, dict):
                return MultiMapsList.from_dict(multi_maps_list)
        except RoborockTimeout as e:
            _LOGGER.error(e)

    async def get_networking(self, device_id) -> NetworkInfo:
        try:
            networking_info = await self.send_command(device_id, RoborockCommand.GET_NETWORK_INFO)
            if isinstance(networking_info, dict):
                return NetworkInfo.from_dict(networking_info)
        except RoborockTimeout as e:
            _LOGGER.error(e)


class RoborockApiClient:
    def __init__(self, username: str, base_url=None) -> None:
        """Sample API Client."""
        self._username = username
        self._default_url = "https://euiot.roborock.com"
        self.base_url = base_url
        self._device_identifier = secrets.token_urlsafe(16)

    async def _get_base_url(self) -> str:
        if not self.base_url:
            url_request = PreparedRequest(self._default_url)
            response = await url_request.request(
                "post",
                "/api/v1/getUrlByEmail",
                params={"email": self._username, "needtwostepauth": "false"},
            )
            if response.get("code") != 200:
                raise RoborockException(response.get("error"))
            self.base_url = response.get("data").get("url")
        return self.base_url

    def _get_header_client_id(self):
        md5 = hashlib.md5()
        md5.update(self._username.encode())
        md5.update(self._device_identifier.encode())
        return base64.b64encode(md5.digest()).decode()

    async def request_code(self) -> None:
        base_url = await self._get_base_url()
        header_clientid = self._get_header_client_id()
        code_request = PreparedRequest(base_url, {"header_clientid": header_clientid})

        code_response = await code_request.request(
            "post",
            "/api/v1/sendEmailCode",
            params={
                "username": self._username,
                "type": "auth",
            },
        )

        if code_response.get("code") != 200:
            raise RoborockException(code_response.get("msg"))

    async def pass_login(self, password: str) -> UserData:
        base_url = await self._get_base_url()
        header_clientid = self._get_header_client_id()

        login_request = PreparedRequest(base_url, {"header_clientid": header_clientid})
        login_response = await login_request.request(
            "post",
            "/api/v1/login",
            params={
                "username": self._username,
                "password": password,
                "needtwostepauth": "false",
            },
        )

        if login_response.get("code") != 200:
            raise RoborockException(login_response.get("msg"))
        return UserData.from_dict(login_response.get("data"))

    async def code_login(self, code) -> UserData:
        base_url = await self._get_base_url()
        header_clientid = self._get_header_client_id()

        login_request = PreparedRequest(base_url, {"header_clientid": header_clientid})
        login_response = await login_request.request(
            "post",
            "/api/v1/loginWithCode",
            params={
                "username": self._username,
                "verifycode": code,
                "verifycodetype": "AUTH_EMAIL_CODE",
            },
        )

        if login_response.get("code") != 200:
            raise RoborockException(login_response.get("msg"))
        return UserData.from_dict(login_response.get("data"))

    async def get_home_data(self, user_data: UserData) -> HomeData:
        base_url = await self._get_base_url()
        header_clientid = self._get_header_client_id()
        rriot = user_data.rriot
        home_id_request = PreparedRequest(
            base_url, {"header_clientid": header_clientid}
        )
        home_id_response = await home_id_request.request(
            "get",
            "/api/v1/getHomeDetail",
            headers={"Authorization": user_data.token},
        )
        if home_id_response.get("code") != 200:
            raise RoborockException(home_id_response.get("msg"))
        home_id = home_id_response.get("data").get("rrHomeId")
        timestamp = math.floor(time.time())
        nonce = secrets.token_urlsafe(6)
        prestr = ":".join(
            [
                rriot.u,
                rriot.s,
                nonce,
                str(timestamp),
                hashlib.md5(("/user/homes/" + str(home_id)).encode()).hexdigest(),
                "",
                "",
            ]
        )
        mac = base64.b64encode(
            hmac.new(rriot.h.encode(), prestr.encode(), hashlib.sha256).digest()
        ).decode()
        home_request = PreparedRequest(
            rriot.r.a,
            {
                "Authorization": f'Hawk id="{rriot.u}", s="{rriot.s}", ts="{timestamp}", nonce="{nonce}", '
                                 f'mac="{mac}"',
            },
        )
        home_response = await home_request.request("get", "/user/homes/" + str(home_id))
        if not home_response.get("success"):
            raise RoborockException(home_response)
        home_data = home_response.get("result")
        return HomeData.from_dict(home_data)



from __future__ import annotations

from enum import Enum


class RoborockEnum(str, Enum):

    def __str__(self):
        return str(self.value)

    @classmethod
    def _missing_(cls, code: int):
        return cls._member_map_.get(str(code))

    @classmethod
    def as_dict(cls):
        return {int(i.name): i.value for i in cls}

    @classmethod
    def values(cls):
        return list(cls.as_dict().values())

    @classmethod
    def keys(cls):
        return list(cls.as_dict().keys())

    @classmethod
    def items(cls):
        return cls.as_dict().items()


class RoborockCode:

    def __new__(cls, name: str, data: dict):
        return RoborockEnum(name, {str(key): value for key, value in data.items()})


RoborockStateCode = RoborockCode("RoborockStateCode", {
    1: "starting",
    2: "charger_disconnected",
    3: "idle",
    4: "remote_control_active",
    5: "cleaning",
    6: "returning_home",
    7: "manual_mode",
    8: "charging",
    9: "charging_problem",
    10: "paused",
    11: "spot_cleaning",
    12: "error",
    13: "shutting_down",
    14: "updating",
    15: "docking",
    16: "going_to_target",
    17: "zoned_cleaning",
    18: "segment_cleaning",
    22: "emptying_the_bin",  # on s7+, see #1189
    23: "washing_the_mop",  # on a46, #1435
    26: "going_to_wash_the_mop",  # on a46, #1435
    100: "charging_complete",
    101: "device_offline",
})

RoborockErrorCode = RoborockCode("RoborockErrorCode", {
    0: "none",
    1: "lidar_blocked",
    2: "bumper_stuck",
    3: "wheels_suspended",
    4: "cliff_sensor_error",
    5: "main_brush_jammed",
    6: "side_brush_jammed",
    7: "wheels_jammed",
    8: "robot_trapped",
    9: "no_dustbin",
    12: "low_battery",
    13: "charging_error",
    14: "battery_error",
    15: "wall_sensor_dirty",
    16: "robot_tilted",
    17: "side_brush_error",
    18: "fan_error",
    21: "vertical_bumper_pressed",
    22: "dock_locator_error",
    23: "return_to_dock_fail",
    24: "no-go_zone_detected",
    27: "vibrarise_jammed",
    28: "robot_on_carpet",
    29: "filter_blocked",
    30: "invisible_wall_detected",
    31: "cannot_cross_carpet",
    32: "internal_error"
})

RoborockFanPowerCode = RoborockCode("RoborockFanPowerCode", {
    105: "off",
    101: "silent",
    102: "balanced",
    103: "turbo",
    104: "max",
    108: "max_plus",
    106: "custom",
})

RoborockMopModeCode = RoborockCode("RoborockMopModeCode", {
    300: "standard",
    301: "deep",
    303: "deep_plus",
    302: "custom",
})

RoborockMopIntensityCode = RoborockCode("RoborockMopIntensityCode", {
    200: "off",
    201: "mild",
    202: "moderate",
    203: "intense",
    204: "custom",
})

RoborockDockErrorCode = RoborockCode("RoborockDockErrorCode", {
    0: "ok",
    38: 'water empty',
    39: 'waste water tank full',
})

RoborockDockTypeCode = RoborockCode("RoborockDockTypeCode", {
    0: "no_dock",
    3: "empty_wash_fill_dock",
})

RoborockDockDustCollectionModeCode = RoborockCode("RoborockDockDustCollectionModeCode", {
    0: "smart",
    1: "light",
    2: "balanced",
    4: "max",
})

RoborockDockWashTowelModeCode = RoborockCode("RoborockDockWashTowelModeCode", {
    0: "light",
    1: "balanced",
    2: "deep",
})



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum

from .containers import Status, CleanSummary, Consumable, \
    DNDTimer, CleanRecord, SmartWashParams, DustCollectionMode, WashTowelMode


class RoborockDevicePropField(str, Enum):
    STATUS = "status"
    DND_TIMER = "dnd_timer"
    CLEAN_SUMMARY = "clean_summary"
    CONSUMABLE = "consumable"
    LAST_CLEAN_RECORD = "last_clean_record"
    DOCK_SUMMARY = "dock_summary"


class RoborockDockSummaryField(str, Enum):
    DUST_COLLECTION_MODE = "dust_collection_mode"
    WASHING_MODE_TYPE = "washing_mode_type"
    MOP_WASH = "mop_wash"


class RoborockCommand(str, Enum):
    GET_PROP = "get_prop"
    GET_MAP_V1 = "get_map_v1"
    GET_STATUS = "get_status"
    GET_DND_TIMER = "get_dnd_timer"
    GET_CLEAN_SUMMARY = "get_clean_summary"
    GET_CLEAN_RECORD = "get_clean_record"
    GET_CONSUMABLE = "get_consumable"
    GET_MULTI_MAPS_LIST = "get_multi_maps_list"
    APP_START = "app_start"
    APP_PAUSE = "app_pause"
    APP_STOP = "app_stop"
    APP_CHARGE = "app_charge"
    APP_SPOT = "app_spot"
    FIND_ME = "find_me"
    RESUME_ZONED_CLEAN = "resume_zoned_clean"
    RESUME_SEGMENT_CLEAN = "resume_segment_clean"
    SET_CUSTOM_MODE = "set_custom_mode"
    SET_MOP_MODE = "set_mop_mode"
    SET_WATER_BOX_CUSTOM_MODE = "set_water_box_custom_mode"
    RESET_CONSUMABLE = "reset_consumable"
    LOAD_MULTI_MAP = "load_multi_map"
    APP_RC_START = "app_rc_start"
    APP_RC_END = "app_rc_end"
    APP_RC_MOVE = "app_rc_move"
    APP_GOTO_TARGET = "app_goto_target"
    APP_SEGMENT_CLEAN = "app_segment_clean"
    APP_ZONED_CLEAN = "app_zoned_clean"
    APP_GET_DRYER_SETTING = "app_get_dryer_setting"
    APP_SET_DRYER_SETTING = "app_set_dryer_setting"
    APP_START_WASH = "app_start_wash"
    APP_STOP_WASH = "app_stop_wash"
    GET_DUST_COLLECTION_MODE = "get_dust_collection_mode"
    SET_DUST_COLLECTION_MODE = "set_dust_collection_mode"
    GET_SMART_WASH_PARAMS = "get_smart_wash_params"
    SET_SMART_WASH_PARAMS = "set_smart_wash_params"
    GET_WASH_TOWEL_MODE = "get_wash_towel_mode"
    SET_WASH_TOWEL_MODE = "set_wash_towel_mode"
    SET_CHILD_LOCK_STATUS = "set_child_lock_status"
    GET_CHILD_LOCK_STATUS = "get_child_lock_status"
    START_WASH_THEN_CHARGE = "start_wash_then_charge"
    GET_CURRENT_SOUND = "get_current_sound"
    GET_SERIAL_NUMBER = "get_serial_number"
    GET_TIMEZONE = "get_timezone"
    GET_SERVER_TIMER = "get_server_timer"
    GET_CUSTOMIZE_CLEAN_MODE = "get_customize_clean_mode"
    GET_CLEAN_SEQUENCE = "get_clean_sequence"
    SET_FDS_ENDPOINT = "set_fds_endpoint"  # Seems to be logging server
    ENABLE_LOG_UPLOAD = "enable_log_upload"
    APP_WAKEUP_ROBOT = "app_wakeup_robot"
    GET_LED_STATUS = "get_led_status"
    GET_FLOW_LED_STATUS = "get_flow_led_status"
    SET_FLOW_LED_STATUS = "set_flow_led_status"
    GET_SOUND_PROGRESS = "get_sound_progress"
    GET_SOUND_VOLUME = "get_sound_volume"
    TEST_SOUND_VOLUME = "test_sound_volume"
    CHANGE_SOUND_VOLUME = "change_sound_volume"
    GET_CARPET_MODE = "get_carpet_mode"
    SET_CARPET_MODE = "set_carpet_mode"
    GET_CARPET_CLEAN_MODE = "get_carpet_clean_mode"
    SET_CARPET_CLEAN_MODE = "set_carpet_clean_mode"
    UPD_SERVER_TIMER = "upd_server_timer"  # Server timer seems to be with schedules
    SET_SERVER_TIMER = "set_server_timer"
    SET_APP_TIMEZONE = "set_app_timezone"
    GET_NETWORK_INFO = "get_network_info"
    GET_IDENTIFY_FURNITURE_STATUS = "get_identify_furniture_status"
    SET_CAMERA_STATUS = "set_camera_status"
    SET_DND_TIMER = "set_dnd_timer"
    GET_COLLISION_AVOID_STATUS = "get_collision_avoid_status"
    CLOSE_VALLEY_ELECTRICITY_TIMER = "close_valley_electricity_timer"
    GET_VALLEY_ELECTRICITY_TIMER = "get_valley_electricity_timer"
    SET_CLEAN_MOTOR_MODE = "set_clean_motor_mode"
    SET_LED_STATUS = "set_led_status"
    GET_CAMERA_STATUS = "get_camera_status"
    CLOSE_DND_TIMER = "close_dnd_timer"
    GET_MULTI_MAP = "get_multi_map"
    SET_COLLISION_AVOID_STATUS = "set_collision_avoid_status"
    SET_IDENTIFY_GROUND_MATERIAL_STATUS = "set_identify_ground_material_status"
    GET_IDENTIFY_GROUND_MATERIAL_STATUS = "get_identify_ground_material_status"
    SET_VALLEY_ELECTRICITY_TIMER = "set_valley_electricity_timer"
    SWITCH_WATER_MARK = "switch_water_mark"
    SET_IDENTIFY_FURNITURE_STATUS = "set_identify_furniture_status"
    GET_CLEAN_RECORD_MAP = "get_clean_record_map"


@dataclass
class CommandInfo:
    prefix: bytes


CommandInfoMap: dict[RoborockCommand, CommandInfo] = {
    RoborockCommand.GET_PROP: CommandInfo(prefix=b'\x00\x00\x00\x87'),
    RoborockCommand.GET_STATUS: CommandInfo(prefix=b'\x00\x00\x00\x77'),
    RoborockCommand.SET_CUSTOM_MODE: CommandInfo(prefix=b'\x00\x00\x00\x87'),
    RoborockCommand.GET_CHILD_LOCK_STATUS: CommandInfo(prefix=b'\x00\x00\x00\x87'),
    RoborockCommand.GET_MULTI_MAPS_LIST: CommandInfo(prefix=b'\x00\x00\x00\x87'),
    RoborockCommand.GET_IDENTIFY_FURNITURE_STATUS: CommandInfo(prefix=b'\x00\x00\x00\x87'),
    RoborockCommand.SET_WATER_BOX_CUSTOM_MODE: CommandInfo(prefix=b'\x00\x00\x00\x87'),
    RoborockCommand.GET_CLEAN_SEQUENCE: CommandInfo(prefix=b'\x00\x00\x00\x87'),
    RoborockCommand.GET_CUSTOMIZE_CLEAN_MODE: CommandInfo(prefix=b'\x00\x00\x00\x87'),
    RoborockCommand.GET_CARPET_CLEAN_MODE: CommandInfo(prefix=b'\x00\x00\x00\x87'),
    RoborockCommand.SET_CAMERA_STATUS: CommandInfo(prefix=b'\x00\x00\x00\x87'),
    RoborockCommand.SET_DND_TIMER: CommandInfo(prefix=b'\x00\x00\x00\x87'),
    RoborockCommand.GET_FLOW_LED_STATUS: CommandInfo(prefix=b'\x00\x00\x00\x87'),
    RoborockCommand.GET_COLLISION_AVOID_STATUS: CommandInfo(prefix=b'\x00\x00\x00\x87'),
    RoborockCommand.CLOSE_VALLEY_ELECTRICITY_TIMER: CommandInfo(prefix=b'\x00\x00\x00\x87'),
    RoborockCommand.SET_FLOW_LED_STATUS: CommandInfo(prefix=b'\x00\x00\x00\x87'),
    RoborockCommand.GET_VALLEY_ELECTRICITY_TIMER: CommandInfo(prefix=b'\x00\x00\x00\x87'),
    RoborockCommand.GET_CLEAN_RECORD: CommandInfo(prefix=b'\x00\x00\x00\x87'),
    RoborockCommand.GET_MAP_V1: CommandInfo(prefix=b'\x00\x00\x00\xc7'),
    RoborockCommand.SET_CLEAN_MOTOR_MODE: CommandInfo(prefix=b'\x00\x00\x00\xb7'),
    RoborockCommand.GET_CONSUMABLE: CommandInfo(prefix=b'\x00\x00\x00w'),
    RoborockCommand.GET_SERVER_TIMER: CommandInfo(prefix=b'\x00\x00\x00w'),
    RoborockCommand.GET_SERIAL_NUMBER: CommandInfo(prefix=b'\x00\x00\x00w'),
    RoborockCommand.GET_CURRENT_SOUND: CommandInfo(prefix=b'\x00\x00\x00w'),
    RoborockCommand.SET_LED_STATUS: CommandInfo(prefix=b'\x00\x00\x00w'),
    RoborockCommand.GET_CAMERA_STATUS: CommandInfo(prefix=b'\x00\x00\x00w'),
    RoborockCommand.APP_PAUSE: CommandInfo(prefix=b'\x00\x00\x00w'),
    RoborockCommand.GET_CLEAN_SUMMARY: CommandInfo(prefix=b'\x00\x00\x00w'),
    RoborockCommand.GET_NETWORK_INFO: CommandInfo(prefix=b'\x00\x00\x00w'),
    RoborockCommand.GET_LED_STATUS: CommandInfo(prefix=b'\x00\x00\x00w'),
    RoborockCommand.CLOSE_DND_TIMER: CommandInfo(prefix=b'\x00\x00\x00w'),
    RoborockCommand.APP_WAKEUP_ROBOT: CommandInfo(prefix=b'\x00\x00\x00w'),
    RoborockCommand.SET_MOP_MODE: CommandInfo(prefix=b'\x00\x00\x00w'),
    RoborockCommand.GET_DND_TIMER: CommandInfo(prefix=b'\x00\x00\x00w'),
    RoborockCommand.GET_CARPET_MODE: CommandInfo(prefix=b'\x00\x00\x00w'),
    RoborockCommand.GET_TIMEZONE: CommandInfo(prefix=b'\x00\x00\x00w'),
    RoborockCommand.SET_CARPET_MODE: CommandInfo(prefix=b'\x00\x00\x00\xd7'),
    RoborockCommand.GET_MULTI_MAP: CommandInfo(prefix=b'\x00\x00\x00\xd7'),
    RoborockCommand.SET_COLLISION_AVOID_STATUS: CommandInfo(prefix=b'\x00\x00\x97'),
    RoborockCommand.SET_CARPET_CLEAN_MODE: CommandInfo(prefix=b'\x00\x00\x97'),
    RoborockCommand.SET_IDENTIFY_GROUND_MATERIAL_STATUS: CommandInfo(prefix=b'\x00\x00\x97'),
    RoborockCommand.GET_IDENTIFY_GROUND_MATERIAL_STATUS: CommandInfo(prefix=b'\x00\x00\x97'),
    RoborockCommand.SET_VALLEY_ELECTRICITY_TIMER: CommandInfo(prefix=b'\x00\x00\x97'),
    RoborockCommand.SWITCH_WATER_MARK: CommandInfo(prefix=b'\x00\x00\x97'),
    RoborockCommand.SET_IDENTIFY_FURNITURE_STATUS: CommandInfo(prefix=b'\x00\x00\x97'),
    RoborockCommand.SET_CHILD_LOCK_STATUS: CommandInfo(prefix=b'\x00\x00\x97'),
    RoborockCommand.GET_CLEAN_RECORD_MAP: CommandInfo(prefix=b'\x00\x00\xe7'),
    RoborockCommand.APP_START: CommandInfo(prefix=b'\x00\x00\x00w'),
    RoborockCommand.APP_STOP: CommandInfo(prefix=b'\x00\x00\x00w'),
    RoborockCommand.APP_CHARGE: CommandInfo(prefix=b'\x00\x00\x00w'),
    RoborockCommand.APP_SPOT: CommandInfo(prefix=b'\x00\x00\x00w'),
    RoborockCommand.FIND_ME: CommandInfo(prefix=b'\x00\x00\x00w'),
    RoborockCommand.RESUME_ZONED_CLEAN: CommandInfo(prefix=b'\x00\x00\x00w'),
    RoborockCommand.RESUME_SEGMENT_CLEAN: CommandInfo(prefix=b'\x00\x00\x00w'),
    RoborockCommand.RESET_CONSUMABLE: CommandInfo(prefix=b'\x00\x00\x00w'),
    RoborockCommand.LOAD_MULTI_MAP: CommandInfo(prefix=b'\x00\x00\x00w'),
    RoborockCommand.APP_RC_START: CommandInfo(prefix=b'\x00\x00\x00w'),
    RoborockCommand.APP_RC_END: CommandInfo(prefix=b'\x00\x00\x00w'),
    RoborockCommand.APP_RC_MOVE: CommandInfo(prefix=b'\x00\x00\x00w'),
    RoborockCommand.APP_GOTO_TARGET: CommandInfo(prefix=b'\x00\x00\x00w'),
    RoborockCommand.APP_SEGMENT_CLEAN: CommandInfo(prefix=b'\x00\x00\x00w'),
    RoborockCommand.APP_ZONED_CLEAN: CommandInfo(prefix=b'\x00\x00\x00w'),
    RoborockCommand.APP_START_WASH: CommandInfo(prefix=b'\x00\x00\x00w'),
    RoborockCommand.APP_STOP_WASH: CommandInfo(prefix=b'\x00\x00\x00w'),
    RoborockCommand.SET_FDS_ENDPOINT: CommandInfo(prefix=b'\x00\x00\x97'),
    RoborockCommand.ENABLE_LOG_UPLOAD: CommandInfo(prefix=b'\x00\x00\x87'),
    RoborockCommand.GET_SOUND_VOLUME: CommandInfo(prefix=b'\x00\x00\x00w'),
    RoborockCommand.TEST_SOUND_VOLUME: CommandInfo(prefix=b'\x00\x00\x00w'),
    RoborockCommand.UPD_SERVER_TIMER: CommandInfo(prefix=b'\x00\x00\x00w'),
    RoborockCommand.SET_APP_TIMEZONE: CommandInfo(prefix=b'\x00\x00\x97'),
    #TODO discover prefix for following commands
    # RoborockCommand.APP_GET_DRYER_SETTING: CommandInfo(prefix=b'\x00\x00\x00w'),
    # RoborockCommand.APP_SET_DRYER_SETTING: CommandInfo(prefix=b'\x00\x00\x00w'),
    # RoborockCommand.GET_DUST_COLLECTION_MODE: CommandInfo(prefix=b'\x00\x00\x00w'),
    # RoborockCommand.SET_DUST_COLLECTION_MODE: CommandInfo(prefix=b'\x00\x00\x00w'),
    # RoborockCommand.GET_SMART_WASH_PARAMS: CommandInfo(prefix=b'\x00\x00\x00w'),
    # RoborockCommand.SET_SMART_WASH_PARAMS: CommandInfo(prefix=b'\x00\x00\x00w'),
    # RoborockCommand.GET_WASH_TOWEL_MODE: CommandInfo(prefix=b'\x00\x00\x00w'),
    # RoborockCommand.SET_WASH_TOWEL_MODE: CommandInfo(prefix=b'\x00\x00\x00w'),
    # RoborockCommand.START_WASH_THEN_CHARGE: CommandInfo(prefix=b'\x00\x00\x00w'),
    # RoborockCommand.GET_SOUND_PROGRESS: CommandInfo(prefix=b'\x00\x00\x00w'),
    # RoborockCommand.CHANGE_SOUND_VOLUME: CommandInfo(prefix=b'\x00\x00\x00w'),
    # RoborockCommand.SET_SERVER_TIMER: CommandInfo(prefix=b'\x00\x00\x00w'),
}


class RoborockDockSummary:
    def __init__(self, dust_collection_mode: DustCollectionMode,
                 wash_towel_mode: WashTowelMode, smart_wash_params: SmartWashParams) -> None:
        self.dust_collection_mode = dust_collection_mode
        self.wash_towel_mode = wash_towel_mode
        self.smart_wash_params = smart_wash_params


class RoborockDeviceProp:
    def __init__(self, status: Status, dnd_timer: DNDTimer, clean_summary: CleanSummary, consumable: Consumable,
                 last_clean_record: CleanRecord = None, dock_summary: RoborockDockSummary = None):
        self.status = status
        self.dnd_timer = dnd_timer
        self.clean_summary = clean_summary
        self.consumable = consumable
        self.last_clean_record = last_clean_record
        self.dock_summary = dock_summary

    def update(self, device_prop: 'RoborockDeviceProp'):
        if device_prop.status:
            self.status = device_prop.status
        if device_prop.dnd_timer:
            self.dnd_timer = device_prop.dnd_timer
        if device_prop.clean_summary:
            self.clean_summary = device_prop.clean_summary
        if device_prop.consumable:
            self.consumable = device_prop.consumable
        if device_prop.last_clean_record:
            self.last_clean_record = device_prop.last_clean_record
        if device_prop.dock_summary:
            self.dock_summary = device_prop.dock_summary



"""Roborock API."""

from roborock.api import RoborockApiClient
from roborock.cloud_api import RoborockMqttClient
from roborock.local_api import RoborockLocalClient
from roborock.containers import *
from roborock.exceptions import *
from roborock.typing import *
from roborock.code_mappings import *



from __future__ import annotations

import base64
import logging
import secrets
import threading
from asyncio import Lock
from typing import Any
from urllib.parse import urlparse

import paho.mqtt.client as mqtt

from roborock.api import md5hex, RoborockClient, SPECIAL_COMMANDS
from roborock.exceptions import (
    RoborockException,
    CommandVacuumError,
    VacuumError,
)
from .containers import (
    UserData,
    RoborockDeviceInfo,
)
from .roborock_message import RoborockParser, md5bin, RoborockMessage
from .roborock_queue import RoborockQueue
from .typing import RoborockCommand
from .util import run_in_executor

_LOGGER = logging.getLogger(__name__)
QUEUE_TIMEOUT = 4
MQTT_KEEPALIVE = 60


class RoborockMqttClient(RoborockClient, mqtt.Client):
    _thread: threading.Thread

    def __init__(self, user_data: UserData, devices_info: dict[str, RoborockDeviceInfo]) -> None:
        rriot = user_data.rriot
        endpoint = base64.b64encode(md5bin(rriot.k)[8:14]).decode()
        RoborockClient.__init__(self, endpoint, devices_info)
        mqtt.Client.__init__(self, protocol=mqtt.MQTTv5)
        self._mqtt_user = rriot.u
        self._hashed_user = md5hex(self._mqtt_user + ":" + rriot.k)[2:10]
        url = urlparse(rriot.r.m)
        self._mqtt_host = url.hostname
        self._mqtt_port = url.port
        self._mqtt_ssl = url.scheme == "ssl"
        if self._mqtt_ssl:
            super().tls_set()
        self._mqtt_password = rriot.s
        self._hashed_password = md5hex(self._mqtt_password + ":" + rriot.k)[16:]
        super().username_pw_set(self._hashed_user, self._hashed_password)
        self._seq = 1
        self._random = 4711
        self._id_counter = 2
        self._salt = "TXdfu$jyZ#TZHsg4"
        self._endpoint = base64.b64encode(md5bin(rriot.k)[8:14]).decode()
        self._nonce = secrets.token_bytes(16)
        self._waiting_queue: dict[int, RoborockQueue] = {}
        self._mutex = Lock()
        self._last_device_msg_in = mqtt.time_func()
        self._last_disconnection = mqtt.time_func()

    def __del__(self) -> None:
        self.sync_disconnect()

    @run_in_executor()
    async def on_connect(self, _client, _, __, rc, ___=None) -> None:
        connection_queue = self._waiting_queue.get(0)
        if rc != mqtt.MQTT_ERR_SUCCESS:
            message = f"Failed to connect (rc: {rc})"
            _LOGGER.error(message)
            if connection_queue:
                await connection_queue.async_put(
                    (None, VacuumError(rc, message)), timeout=QUEUE_TIMEOUT
                )
            return
        _LOGGER.info(f"Connected to mqtt {self._mqtt_host}:{self._mqtt_port}")
        topic = f"rr/m/o/{self._mqtt_user}/{self._hashed_user}/#"
        (result, mid) = self.subscribe(topic)
        if result != 0:
            message = f"Failed to subscribe (rc: {result})"
            _LOGGER.error(message)
            if connection_queue:
                await connection_queue.async_put(
                    (None, VacuumError(rc, message)), timeout=QUEUE_TIMEOUT
                )
            return
        _LOGGER.info(f"Subscribed to topic {topic}")
        if connection_queue:
            await connection_queue.async_put((True, None), timeout=QUEUE_TIMEOUT)

    @run_in_executor()
    async def on_message(self, _client, _, msg, __=None) -> None:
        async with self._mutex:
            self._last_device_msg_in = mqtt.time_func()
        device_id = msg.topic.split("/").pop()
        messages, _ = RoborockParser.decode(msg.payload, self.devices_info[device_id].device.local_key)
        await super().on_message(messages)

    @run_in_executor()
    async def on_disconnect(self, _client: mqtt.Client, _, rc, __=None) -> None:
        try:
            self._last_disconnection = mqtt.time_func()
            message = f"Roborock mqtt client disconnected (rc: {rc})"
            _LOGGER.warning(message)
            connection_queue = self._waiting_queue.get(1)
            if connection_queue:
                await connection_queue.async_put(
                    (True, None), timeout=QUEUE_TIMEOUT
                )
        except Exception as ex:
            _LOGGER.exception(ex)

    @run_in_executor()
    async def _async_check_keepalive(self) -> None:
        async with self._mutex:
            now = mqtt.time_func()
            if now - self._last_disconnection > self._keepalive ** 2 and now - self._last_device_msg_in > self._keepalive:
                self._ping_t = self._last_device_msg_in

    def _check_keepalive(self) -> None:
        self._async_check_keepalive()
        super()._check_keepalive()

    def sync_stop_loop(self) -> None:
        if self._thread:
            _LOGGER.info("Stopping mqtt loop")
            super().loop_stop()

    def sync_start_loop(self) -> None:
        if not self._thread or not self._thread.is_alive():
            self.sync_stop_loop()
            _LOGGER.info("Starting mqtt loop")
            super().loop_start()

    def sync_disconnect(self) -> bool:
        rc = mqtt.MQTT_ERR_AGAIN
        if self.is_connected():
            _LOGGER.info("Disconnecting from mqtt")
            rc = super().disconnect()
            if not rc in [mqtt.MQTT_ERR_SUCCESS, mqtt.MQTT_ERR_NO_CONN]:
                raise RoborockException(f"Failed to disconnect (rc:{rc})")
        return rc == mqtt.MQTT_ERR_SUCCESS

    def sync_connect(self) -> bool:
        rc = mqtt.MQTT_ERR_AGAIN
        self.sync_start_loop()
        if not self.is_connected():
            _LOGGER.info("Connecting to mqtt")
            rc = super().connect(
                host=self._mqtt_host,
                port=self._mqtt_port,
                keepalive=MQTT_KEEPALIVE
            )
            if rc != mqtt.MQTT_ERR_SUCCESS:
                raise RoborockException(f"Failed to connect (rc:{rc})")
        return rc == mqtt.MQTT_ERR_SUCCESS

    async def async_disconnect(self) -> Any:
        async with self._mutex:
            disconnecting = self.sync_disconnect()
            if disconnecting:
                (response, err) = await self._async_response(1)
                if err:
                    raise RoborockException(err) from err
                return response

    async def async_connect(self) -> Any:
        async with self._mutex:
            connecting = self.sync_connect()
            if connecting:
                (response, err) = await self._async_response(0)
                if err:
                    raise RoborockException(err) from err
                return response

    async def validate_connection(self) -> None:
        await self.async_connect()

    def _send_msg_raw(self, device_id, msg) -> None:
        info = self.publish(
            f"rr/m/i/{self._mqtt_user}/{self._hashed_user}/{device_id}", msg
        )
        if info.rc != mqtt.MQTT_ERR_SUCCESS:
            raise RoborockException(f"Failed to publish (rc: {info.rc})")

    async def send_command(
            self, device_id: str, method: RoborockCommand, params: list = None
    ):
        await self.validate_connection()
        request_id, timestamp, payload = super()._get_payload(method, params, True)
        _LOGGER.debug(f"id={request_id} Requesting method {method} with {params}")
        request_protocol = 101
        response_protocol = 301 if method in SPECIAL_COMMANDS else 102
        roborock_message = RoborockMessage(
            timestamp=timestamp,
            protocol=request_protocol,
            payload=payload
        )
        local_key = self.devices_info[device_id].device.local_key
        msg = RoborockParser.encode(roborock_message, local_key)
        self._send_msg_raw(device_id, msg)
        (response, err) = await self._async_response(request_id, response_protocol)
        if err:
            raise CommandVacuumError(method, err) from err
        if response_protocol == 301:
            _LOGGER.debug(
                f"id={request_id} Response from {method}: {len(response)} bytes"
            )
        else:
            _LOGGER.debug(f"id={request_id} Response from {method}: {response}")
        return response

    async def get_map_v1(self, device_id):
        return await self.send_command(device_id, RoborockCommand.GET_MAP_V1)



"""Roborock exceptions."""

class RoborockException(BaseException):
    """Class for Roborock exceptions."""

class RoborockTimeout(RoborockException):
    """Class for Roborock timeout exceptions."""

class RoborockConnectionException(RoborockException):
    """Class for Roborock connection exceptions."""

class RoborockBackoffException(RoborockException):
    """Class for Roborock exceptions when many retries were made."""

class VacuumError(RoborockException):
    """Class for vacuum errors."""
    def __init__(self, code, message):
        self.code = code
        self.message = message
        super().__init__()

    def __str__(self, *args, **kwargs): # real signature unknown
        """ Return str(self). """
        return f"{self.code}: {self.message}"


class CommandVacuumError(RoborockException):
    """Class for command vacuum errors."""
    def __init__(self, command: str, vacuum_error: VacuumError):
        self.message = f"{command}: {str(vacuum_error)}"
        super().__init__(self.message)
